/*
Copyright  2004-2006 Eland Systems All Rights Reserved.

   1. Redistribution and use in source and binary forms must retain the above
   copyright notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by Eland Systems

   4. The name of Eland Systems may not be used to endorse or promote products
   derived from this software without specific prior written permission.

   Author scam+sshlockd@elandsys.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sysexits.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <sys/time.h>

#include <syslog.h>
#include <dirent.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <regex.h>

#ifndef LINUX
#include <sys/queue.h>
#else
#include "queue.h"
#endif

#define VERSION "1.1"
#define SSHLOCKCONF "/etc/sshlockd.conf"

struct CIDR {
	unsigned long ip;
	short int masklen;
	SLIST_ENTRY(CIDR)	cidrs;
};

SLIST_HEAD(, CIDR)	cidrlist;

static volatile u_int sigexit = 0;

regex_t re1;
regex_t re2;

struct entry {
	struct in_addr addr;
	long timemod;
	unsigned long cnt;
	TAILQ_ENTRY(entry)	entries;
};

TAILQ_HEAD(locklist, entry);
struct locklist failures;

int loadre()
{
	int i;
	const char *error;
	int erroffset;
	char    buf[100];

	char regline0[] = "sshd\\[([0-9]+)]: Failed password for invalid user ([A-Za-z0-9]+) from ([0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+) port";
	char regline1[] = "sshd\\[([0-9]+)]: Failed password for ([A-Za-z0-9]+) from ([0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+) port";

	i = 0;
	if((erroffset = regcomp(&re1 , regline0, REG_EXTENDED)) != 0)
	{
		regerror(erroffset, &re1, buf, 120);
		printf("error 3 for line %d %s\n", i, buf);
		return -3;
	}

	i = 1;
	if((erroffset = regcomp(&re2 , regline1, REG_EXTENDED)) != 0)
	{
		printf("error 3 for line %d %s\n", i, buf);
		return -3;
	}

	return 0;
}

int
ip_cidr(const unsigned long ip, const short int masklen, const unsigned long checkip)
{
	unsigned long ipaddr = 0;
	unsigned long cidrip = 0;
	unsigned long subnet = 0;

	subnet = ~0;
	subnet = subnet << (32 - masklen);

	cidrip = htonl(ip) & subnet;
	ipaddr = ntohl(checkip) & subnet;

	if (cidrip == ipaddr)
		return 1;

	return 0;
}

void listlock (const unsigned char *c)
{

	struct entry *ent;
	struct entry *entadd;
	struct entry *entnext;
	int ret = 0;
	struct in_addr addr;
	struct timeval gnow;
	long timemod;
	int retsys = 0;
	char cmd[256];
	struct CIDR *entcidr;
#ifndef LINUX
	struct timezone tz;

	gettimeofday(&gnow, &tz);
#else
	gettimeofday(&gnow, NULL);
#endif

	if (inet_pton(AF_INET, c, &addr) != 1)
		return;

	SLIST_FOREACH(entcidr, &cidrlist, cidrs)
	{
		if (ip_cidr(entcidr->ip, entcidr->masklen, addr.s_addr) == 1)
		{
			syslog (LOG_INFO, "safehost %s", c);
			return;
		}
	}

	timemod =gnow.tv_sec / 1000;

	for (ent = TAILQ_FIRST(&failures); ent; ent = entnext)
	{
		entnext = TAILQ_NEXT(ent, entries);
		if ((ent->addr.s_addr == addr.s_addr) &&  (ent->timemod >= timemod))
		{
			ent->cnt++;
			ret = 1;

			if (ent->cnt > 4)
			{
				/* ip block */
#ifdef LINUX
#ifdef IPCHAINS
				snprintf( cmd, 255, "/sbin/ipchains -I INPUT -p tcp -s %s --dport 22 -j DENY", c);
#else
				snprintf( cmd, 255, "/usr/bin/iptables -I INPUT -p tcp -s %s --dport 22 -j DROP", c);
#endif /* IPCHAINS */
#endif /* LINUX */

#ifndef LINUX
				snprintf( cmd, 255,"/bin/echo 'block in log proto tcp from %s/32 to any port = 22' | /sbin/ipf -f -", c);
#endif

				retsys = system(cmd);
				if ((retsys >= 0) && (retsys < 127))
				{
					syslog (LOG_INFO, "ssh lock %s", c);
				} else
					syslog (LOG_ERR, "ssh cannot lock %s", c);

				TAILQ_REMOVE(&failures, ent, entries);
				return;
			}

		}
		else if (ent->timemod < timemod)
		{
			TAILQ_REMOVE(&failures, ent, entries);
		}
	}

	if (ret == 0)
	{
		entadd = (struct entry *)malloc(sizeof(struct entry));
		if (entadd != NULL)
		{
			entadd->addr.s_addr = addr.s_addr;
			entadd->timemod = timemod;
			entadd->cnt = 1;
			TAILQ_INSERT_TAIL(&failures, entadd, entries);
		}
	}
}

int processlog(const char *logfile, int pos)
{
	FILE *fh;
	char *line;
	int linelen;
	const int clen = 2048;
	int cnt;
	int rc;
	const char **restr;
	int j;

	int linecount = 0;

	struct entry *ent;
	struct entry *entnext;
	regmatch_t pmatch[4];
	char ip[16];

	fh = fopen( logfile, "r");
	if (fh != NULL)
	{
		pos = lseek( fileno(fh), (off_t) pos, SEEK_SET);

		while (!feof(fh))
		{
			if ((line = (char *) malloc(clen)) == NULL)
			{
				fclose(fh);
				return -2;
			}

			memset( line, 0, clen);
			fgets( line, clen - 1, fh);

			/* Assume \n in line */
			linelen = strlen(line);
			if (linelen > 2)
			{
				if (line[linelen - 1] == '\n')
					line[--linelen] = 0;

				if (regexec(&re1, line, 4, pmatch, 0) == 0)
				{
					int size = pmatch[3].rm_eo - pmatch[3].rm_so;
					if (size < 16)
					{
						memcpy( ip, line + pmatch[3].rm_so, size);
						ip[size] = 0;
						listlock(ip);
					}
				} else if (regexec(&re2, line, 4, pmatch, 0) == 0)
				{
					int size = pmatch[3].rm_eo - pmatch[3].rm_so;
					if (size < 16)
					{
						memcpy( ip, line + pmatch[3].rm_so, size);
						ip[size] = 0;
						listlock(ip);
					}
				}

			}
			free(line);
			linecount++;
			if (linecount > 99999)
				break;
		}
		pos = lseek( fileno(fh), (off_t) 0, SEEK_CUR);
		fclose(fh);

	}

	return pos;
}

static void
catch_signal( int signo )
{
	switch (signo) {
		case SIGTERM:
			sigexit = 1;
			break;
	}
}


void
daemonize()
{

	signal (SIGUSR1, catch_signal);
	signal (SIGUSR2, catch_signal);

	(void)close(0);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(1);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(2);
	(void)open("/dev/null", O_RDONLY, 0);

	switch (fork ())
	{
		case 0:
			break;

		case -1:
			perror ("cannot fork\n");
			exit (3);
			break;

		default:
			exit(0);
	}

	if (setsid() < 0)
		exit(1);

	//chdir("/");
}

int
read_conf()
{
	FILE *fh;
	const LINELENGTH = 1024;
	char line[LINELENGTH + 1];
	char buf[LINELENGTH + 1];
	char bufstr[LINELENGTH + 1];
	int lline;
	unsigned int num;
	unsigned long l;
	struct CIDR *entcidr;
	unsigned int ip1 = 0;
	unsigned int ip2 = 0;
	unsigned int ip3 = 0;
	unsigned int ip4 = 0;
	unsigned int ipm = 0;


	if ((fh = fopen( SSHLOCKCONF, "r")) == NULL) {
		return 1;
	}

	syslog (LOG_INFO, "reading configuration");

	while (!SLIST_EMPTY(&cidrlist))
	{
		entcidr = SLIST_FIRST(&cidrlist);
		SLIST_REMOVE_HEAD(&cidrlist, cidrs);
		free(entcidr);
	}

	while( fgets(line, LINELENGTH, fh) )
	{
		if (line[0] == '#')
			continue;
		lline = strlen(line);
		if (lline < 12)
			continue;

		if (sscanf( line, "safeaddr=%u.%u.%u.%u/%u", &ip1, &ip2, &ip3, &ip4, &ipm) == 5)
		{
			snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
			entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
			entcidr->ip = inet_addr(buf);
			entcidr->masklen = ipm;
			SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
		}
		else if (sscanf( line, "safeaddr=%u.%u.%u.%u", &ip1, &ip2, &ip3, &ip4) == 4)
		{
			snprintf( buf, sizeof(buf), "%u.%u.%u.%u", ip1, ip2, ip3, ip4);
			entcidr = (struct CIDR *)malloc(sizeof(struct CIDR));
			entcidr->ip = inet_addr(buf);
			entcidr->masklen = 32;
			SLIST_INSERT_HEAD(&cidrlist, entcidr, cidrs);
		}
	}
	fclose(fh);

	return 0;
}

int main(int argc, char *argv[])
{
	const char *args = "f";
	extern char *optarg;
	int daemonmode = 1;
	int logpos = 0;
	int c;
	int ret;
	struct stat st;
	char proctitle[256];

#ifndef MESSAGES
#ifndef LINUX
	char logname[] = "/var/log/authlog";
#else
	char logname[] = "/var/log/auth";
#endif
#else
	char logname[] = "/var/log/messages";
#endif
	/* Process command line options */
	while ((c = getopt(argc, argv, args)) != (char)EOF)
	{
		switch (c)
		{
		  case 'f':
			  daemonmode = 0;
			  break;

		  case 'h':
		  default:
			exit(0);

		}
	}

	if (loadre() != 0)
		exit(1);

	if (read_conf() != 0)
		exit(2);

	(void) closelog();
	openlog("sshlockd", LOG_PID | LOG_NDELAY, LOG_DAEMON);

	if (daemonmode == 1)
	{
		daemonize();
	}

	if (stat(logname, &st) == 0)
	{
		logpos = st.st_size;
	}

	/* for testing */
	/* logpos = 0; */

	signal (SIGTERM, catch_signal);

	TAILQ_INIT(&failures);

	while (sigexit == 0)
	{
		sleep(20);
		if (stat(logname, &st) == 0)
		{
			if (logpos != st.st_size)
			{
				if (logpos > st.st_size)
					logpos = 0;
#if HAVE_SETPROCTITLE
				snprintf(proctitle, sizeof(proctitle), "reading");
				setproctitle("%s", proctitle);
#endif /* HAVE_SETPROCTITLE */
				ret = processlog(logname, logpos);
				if (ret > 0)
				{
					logpos = ret;
				}
#if HAVE_SETPROCTITLE
		setproctitle(NULL);
#endif /* HAVE_SETPROCTITLE */
			}
		}
	}

	syslog(LOG_INFO, "Exit");
	return 0;
}

